import os
import re


class Image:

    def __init__(self, markdown_link):
        self._original_md_link = markdown_link
        self.updated_md_link = markdown_link

    @property
    def original_md_link(self):
        return self._original_md_link

    @property
    def updated_md_link(self):
        return self._updated_md_link

    @updated_md_link.setter
    def updated_md_link(self, new_link):
        self._updated_md_link = new_link
        self._path = re.search(r'(?<=\().+?(?=\))', self._updated_md_link).group()
        self._file_name = os.path.basename(self._path)

    @property
    def path(self):
        return self._path

    @path.setter
    def path(self, new_path):
        self._path = new_path
        self._updated_md_link = re.sub(r'(?<=\().+(?=\))', new_path, self._updated_md_link)
        self._file_name = os.path.basename(new_path)

    @property
    def file_name(self):
        return self._file_name

    @property
    def abs_path(self):
        return os.path.realpath(self._path)

    def rename(self, new_file_name):
        """
        Rename original image file

        :param str new_file_name: new name of the file (with extension)
        """
        folder = os.path.dirname(self._path)
        new_path = f'{folder}/{new_file_name}'
        os.rename(self._path, new_path)

        self._file_name = new_file_name
        self._path = new_path
        self._updated_md_link = re.sub(r'(?<=\().+(?=\))', new_path, self._updated_md_link)



import mistune


class Converter:

    @classmethod
    def convert_cards(cls, cards_list):
        for card in cards_list:
            cls.convert_card(card)

    @classmethod
    def convert_card(cls, card):
        card.front_converted = cls.convert_string(card.front_raw)
        card.back_converted = cls.convert_string(card.back_raw)

    @classmethod
    def convert_string(cls, raw_str):
        return mistune.html(raw_str)



import filecmp
import os
import random
import re
import shutil
import sys
from typing import List

from .card import Card
from .image import Image

DEFAULT_ANKI_FOLDERS = {
    'win32': r'~\AppData\Roaming\Anki2',
    'linux': '~/.local/share/Anki2',
    'darwin': '~/Library/Application Support/Anki2'
}


class Parser:
    _section_regex = '^---\n(.+?)^---$'
    _deck_name_regex = '(?<=^Deck:).*?$'
    _tags_line_regex = '(?<=^Tags:).*?$'
    _card_substring_regex = r'^\d+\.[\s\S]+?(?:^>.*?(?:\n|$))+'
    _question_regex = r'^\d+\.[\s\S]+?(?=^>)'
    _answer_regex = r'(?:^>.*?(?:\n|$))+'

    def __init__(
            self, file_path: str,
            anki_user_name: str = 'User 1',
    ):
        self._file_path = file_path

        anki_folder_path = os.path.expanduser(DEFAULT_ANKI_FOLDERS[sys.platform])
        self._anki_media_path = f'{anki_folder_path}/{anki_user_name}/collection.media'

    def collect_cards(self) -> List[Card]:
        """Get all cards from the file which path was passed to the Parser"""
        with open(self._file_path, mode='rt', encoding='utf-8') as f:
            file_string = f.read()

        question_sections = self._get_sections(file_string)

        cards = []
        for section in question_sections:
            cards.extend(self._get_cards_from_section(section))

        return cards

    def _get_sections(self, file_contents: str) -> List[str]:
        """Get all sections (groups of cards) from the file string"""
        return re.findall(self._section_regex,
                          file_contents,
                          re.MULTILINE | re.DOTALL)

    def _get_cards_from_section(self, section: str) -> List[Card]:
        """Get all Cards from the section string"""
        section = self._handle_images(section)

        tags = self._get_tags(section)
        deck_name = self._get_deck_name(section)

        # Get all section's substrings which contain question-answer pairs
        card_substrings = self._get_card_substrings(section)

        # Create cards
        cards = []
        for substring in card_substrings:
            question = self._get_question(substring)

            answer = self._get_answer(substring)

            cards.append(Card(front=question,
                              back=answer,
                              tags=tags,
                              deck_name=deck_name))

        return cards

    def _handle_images(self, section: str) -> str:
        """
        Copy images to Anki media folder and change their source
        to be just filename (for Anki to find them)
        """
        # Find all unique images in the text
        image_links = set(re.findall(r'!\[.*?]\(.*?\)', section))
        images = [Image(link) for link in image_links]

        # Copy images to Anki Media folder
        # And change all image links in section string
        for image in images:
            try:
                self._copy_image_to_anki_media(image)
            except OSError:
                raise OSError(f"Couldn't find image '{image.original_md_link}' in path '{image.abs_path}'!")

            # Update all image link occurrences with new one which has updated source
            section = section.replace(image.original_md_link, image.updated_md_link)

        return section

    def _copy_image_to_anki_media(self, image: Image):
        """Copy image to Anki media folder"""
        path_to_anki_image = f'{self._anki_media_path}/{image.file_name}'

        # Check if image already exists in Anki Media folder
        if os.path.exists(path_to_anki_image):
            # If same image is already in folder then skip
            if filecmp.cmp(image.abs_path, path_to_anki_image):
                image.path = image.file_name
                return

            # If not same then rename our image
            image.rename(f'{random.randint(100000, 999999)}_{image.file_name}')
            path_to_anki_image = f'{self._anki_media_path}/{image.file_name}'

        # Copy image
        shutil.copyfile(image.abs_path, path_to_anki_image)

        # Change path to be just file name (for it to work in Anki)
        image.path = image.file_name

    def _get_tags(self, section: str) -> List[str]:
        """Get tags specified for this section"""
        matches = re.findall(self._tags_line_regex,
                             section,
                             re.MULTILINE)
        if not matches:
            return []

        if len(matches) > 1:
            raise ValueError(f'More than one tag field in section:\n{section}')

        tags = matches[0].strip().split()
        return tags

    def _get_deck_name(self, section: str) -> str:
        """Get deck name specified for this section"""
        matches = re.findall(self._deck_name_regex,
                             section,
                             re.MULTILINE)

        if not matches:
            raise ValueError(f"Couldn't find deck name in section:\n{section}")

        if len(matches) > 1:
            raise ValueError(f'More than one deck name field in section:\n{section}')

        deck_name = matches[0].strip()
        if not deck_name:
            raise ValueError(f"Empty deck name field in section:\n{section}")

        return deck_name

    def _get_card_substrings(self, section: str) -> List[str]:
        """Get all section substrings with question-answer pairs"""
        return re.findall(self._card_substring_regex,
                          section,
                          re.MULTILINE)

    def _get_question(self, text: str) -> str:
        """Get clean question string from text
         (without digit followed by period and trailing whitespace)"""
        question_match = re.search(self._question_regex,
                                   text,
                                   re.MULTILINE)

        question = re.sub(r'\d+\.', '', question_match.group(), 1).strip()

        return question

    def _get_answer(self, text: str) -> str:
        """Get clean answer string from text (without '>' and trailing whitespace)"""
        answer_match = re.search(self._answer_regex,
                                 text,
                                 re.MULTILINE)

        # Remove first char ('>') and whitespace at the start
        # and the end of each line
        lines = answer_match.group().splitlines()
        lines = map(lambda l: l[1:].strip(), lines)

        answer = '\n'.join(lines)

        return answer



import requests


class AnkiApi:
    note_type = "Basic"
    front_field_name = "Front"
    back_field_name = "Back"
    api_url = "http://localhost:8765"

    @staticmethod
    def add_cards(cards_list):
        # TODO
        # Use addNotes api call
        for card in cards_list:
            AnkiApi.add_card(card)

        print("All cards sent successfully!")

    @classmethod
    def add_card(cls, card):
        note_dict = cls.create_note_dict(card)
        request_dict = cls.create_base_dict()
        request_dict["action"] = "addNote"
        request_dict["params"]["note"] = note_dict

        # Send request to AnkiConnect
        response = requests.post(cls.api_url, json=request_dict)

        response_json = response.json()

        # Show error message
        if response_json['result'] is None:
            card.print_card_info()
            print("ERROR: Can't create the card!")
            print(f'Reason: "{response_json["error"]}"')
            input("Press Enter to continue...")
            print()

    @classmethod
    def create_note_dict(cls, card):
        return {
            "deckName": card.deck_name,
            "modelName": cls.note_type,
            "fields": {
                cls.front_field_name: card.front_converted,
                cls.back_field_name: card.back_converted
            },
            "options": {
                "allowDuplicate": False,
                "duplicateScope": None,
                "duplicateScopeOptions": {
                    "checkChildren": False
                }
            },
            "tags": card.tags
        }

    @classmethod
    def create_base_dict(cls):
        return {"action": "", "version": 6, "params": {}}



import argparse
import os
import sys

import requests

from .anki_api import AnkiApi
from .converter import Converter
from .parser import Parser


def init_argparse():
    arg_parser = argparse.ArgumentParser(
        prog="inka",
        usage="%(prog)s FILE...",
        description="Extract Anki cards from your Markdown notes."
    )

    arg_parser.add_argument(
        "files",
        metavar="FILE",
        nargs="+",
        type=str,
        help="the path to the file(s)")

    return arg_parser


def create_cards(file_path):
    print(f"Starting to create cards from \"{file_path}\"!")

    abs_file_path = os.path.realpath(file_path)

    # We need to change working directory because images in file can have relative path
    os.chdir(os.path.dirname(abs_file_path))

    note_parser = Parser(abs_file_path)

    print("Getting cards from the file...")
    cards_list = note_parser.collect_cards()

    number_of_cards = len(cards_list)
    if number_of_cards == 0:
        raise ValueError("Cards weren't found in the file.")

    print(f"Found {number_of_cards} cards!")

    print("Converting cards to the html...")
    Converter.convert_cards(cards_list)

    print("Sending cards...")
    try:
        AnkiApi.add_cards(cards_list)
    except requests.exceptions.ConnectionError:
        print("ERROR: Couldn't connect to Anki. "
              "Please, check that Anki is working and you have AnkiConnect plugin is installed.")
        sys.exit()


def main():
    arg_parser = init_argparse()
    args = arg_parser.parse_args()

    original_directory = os.getcwd()
    for file in args.files:
        try:
            create_cards(file)
        except (OSError, ValueError) as e:
            print(e)
            print('Skipping file...')

        os.chdir(original_directory)



class Card:
    def __init__(self, front, back, tags, deck_name):
        self.front_raw = front
        self.back_raw = back
        self.front_converted = None
        self.back_converted = None
        self.tags = tags
        self.deck_name = deck_name

    def print_card_info(self):
        print("------------------------------------")
        print(f"Front: {self.front_raw.strip()}")
        print(f"Back: {self.back_raw.strip()}")
        print(f"Tags: {self.tags}")
        print(f"Deck: {self.deck_name}")
        print("------------------------------------")

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return False

        return (self.front_raw == other.front_raw and
                self.back_raw == other.back_raw and
                self.tags == other.tags and
                self.deck_name == other.deck_name)





from setuptools import setup, find_packages

with open('README.md', mode='rt', encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='inka',
    version='0.8.1',
    author='Salnikov Kirill',
    author_email='salnikov.k54@gmail.com',
    description='Command-line tool for adding Markdown cards to Anki',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/lazy-void/Inka',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Operating System :: OS Independent',
        'Intended Audience :: Education',
        'Topic :: Education :: Computer Aided Instruction (CAI)',
        'Topic :: Text Processing :: Markup :: Markdown',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3 :: Only'
    ],
    license='GPLv3',
    keywords='anki, markdown, spaced-repetition',
    package_dir={'': 'src'},
    packages=find_packages(where='src'),
    python_requires='>=3.6',
    install_requires=[
        'mistune>=2.0.0a6',
        'requests'
    ],
    entry_points={
        'console_scripts': [
            'inka=inka.inka:main'
        ]
    }
)
